import shlex
from collections.abc import AsyncIterable

from pydantic import Field

from ..tracker import Tracker
from ..utils import allow_singular_none
from .base import Package
from .manual_version import ManualVersion
from .system import SystemPackage


class NpmPackage(ManualVersion):
    package: str = Field(..., alias="npm")

    binaries: list[str] = Field(default_factory=list, alias="binary")
    binaries_val = allow_singular_none("binaries")

    def derive_name(self) -> str:
        return self.package

    async def get_remote_version(self) -> str:
        try:
            result = await self.driver.run_output(
                "npm", "view", self.package, "version"
            )
        except Exception as exc:
            raise Exception(
                f"Cannot find latest version of package '{self.package}'."
            ) from exc
        version = result.strip()
        return version

    async def install(self, *, tracker: Tracker) -> None:
        args = ["npm", "exec", "--yes", "--package", self.package, "--"]

        await self.driver.run(*args, "true")

        npx_paths = await self.driver.run_output(*args, "sh", "-c", "echo $PATH")
        npx_path = next((path for path in npx_paths.split(":") if "_npx" in path), None)
        if not npx_path:
            raise Exception(
                f"Could not find npx path in {npx_paths}."
            )  # pragma: no cover

        for name in self.binaries:
            target = await self.driver.local() / "bin" / name
            await self.driver.write_file(
                target,
                f'#!/bin/sh\nPATH={shlex.quote(npx_path)}:$PATH\nexec "{name}" "$@"',
            )
            await self.driver.make_executable(target)
            tracker.track(target, root=self.root)

        await super().install(tracker=tracker)

    async def prerequisites(self) -> AsyncIterable[Package]:
        async for package in super().prerequisites():
            yield package  # pragma: no cover

        os = await self.driver.os()

        for system in os.switch_(
            linux=lambda linux: [
                "nodejs",
                {
                    "debian": "npm",
                    "ubuntu": "npm",
                    "fedora": "nodejs-npm",
                }[linux.distribution],
            ],
            macos=lambda: ["node"],
        ):
            yield SystemPackage(
                system=system,
                db=self.db,
                driver=self.driver_,
            )



from urllib.parse import urlparse

from pydantic import Field

from ..compute import Value, compute
from ..utils import async_cached, url_version
from .archive import ArchivePackage


class URLPackage(ArchivePackage):
    url_: Value = Field(..., alias="url")

    @async_cached
    async def url(self) -> str:
        return await compute(self.url_)

    async def archive_url(self) -> str:
        return await self.url()

    def derive_name(self) -> str:
        if isinstance(self.url_, str):
            url_str = self.url_
        else:
            raise ValueError(
                "Cannot derive name from computed URL and no name override."
            )

        url = urlparse(url_str)

        name = url.hostname or ""

        path = url.path.rsplit("/", 1)[-1]
        if path:
            name += "/" + path.split(".")[0]
        return name

    async def get_remote_version(self) -> str:
        return await url_version(await self.url())



from configparser import ConfigParser as BaseConfigParser
from io import StringIO
from typing import Optional, TypeVar

CP = TypeVar("CP", bound="ConfigParser")


class ConfigParser(BaseConfigParser):
    def optionxform(self, optionstr: str) -> str:
        return optionstr

    def to_string(self) -> str:
        contents = StringIO()
        self.write(contents, space_around_delimiters=False)
        return contents.getvalue()

    @classmethod
    def from_string(cls: type[CP], config: str) -> CP:
        parser = cls(interpolation=None)
        parser.read_string(config)
        return parser


class DesktopEntry(ConfigParser):
    @property
    def icon(self) -> Optional[str]:
        return self["Desktop Entry"].get("Icon")



from collections.abc import AsyncIterable
from pathlib import Path
from typing import Optional

from ..tracker import Tracker
from ..utils import async_cached
from .base import Package
from .system import SystemPackage

SHELLS_FILE = Path("/etc/shells")


class Shell(Package):
    shell: Path

    def derive_name(self) -> str:
        return "_shell"

    async def get_remote_version(self) -> str:
        return str(self.shell)

    async def get_local_version_linux(self) -> str:
        result = await self.driver.with_root(False).run_output(
            "getent", "passwd", await self.driver.username()
        )
        return result.split(":")[6]

    async def get_local_version_macos(self) -> str:
        result = await self.driver.with_root(False).run_output(
            "dscl", ".", "-read", f"/Users/{await self.driver.username()}", "UserShell"
        )
        return result.split(": ")[1]

    async def local_version(self) -> Optional[str]:
        return await (await self.driver.os()).switch(
            linux=self.get_local_version_linux,
            macos=self.get_local_version_macos,
        )()

    @async_cached
    async def all_shells(self) -> list[Path]:
        shells_file = await self.driver.read_file(SHELLS_FILE)
        return [Path(shell) for shell in shells_file.splitlines()]

    async def install(self, *, tracker: Tracker) -> None:
        if not await self.driver.is_file(self.shell):
            raise ValueError(f"{self.shell} does not exist.")

        if not await self.driver.is_executable(self.shell):
            raise ValueError(f"{self.shell} is not executable.")

        if self.shell not in await self.all_shells():
            await self.driver.with_root(True).run(
                "tee", "-a", SHELLS_FILE, input=str(self.shell).encode()
            )

        await self.driver.run("chsh", "-s", self.shell, show_output=True)

        await super().install(tracker=tracker)

    async def prerequisites(self) -> AsyncIterable[Package]:
        async for package in super().prerequisites():
            yield package  # pragma: no cover

        os = await self.driver.os()

        for system in os.switch_(
            linux=lambda linux: (
                ["util-linux-user"] if linux.distribution == "fedora" else []
            ),
            macos=lambda: [],
        ):
            yield SystemPackage(
                system=system,
                db=self.db,
                driver=self.driver_,
            )



import json
import re
from collections.abc import AsyncIterable
from pathlib import Path
from typing import Any, Optional

from pydantic import Field, field_validator

from ..tracker import Tracker
from ..utils import GIT_PREFIX, repo_version
from .base import Package
from .github import GitHubPackage
from .manual_version import ManualVersion
from .system import SystemPackage


class PipxPackage(ManualVersion):
    package: str = Field(..., alias="pipx")

    @field_validator("package")
    @classmethod
    def package_to_lower(cls, value: str) -> str:
        return value.lower()

    def derive_name(self) -> str:
        return self.package

    async def get_metadata(self) -> Optional[dict[str, Any]]:
        pipx_list = json.loads(
            await self.driver.run_output("pipx", "list", "--json", silent=True)
        )
        packages = {
            k: v["metadata"]["main_package"] for k, v in pipx_list["venvs"].items()
        }
        if self.is_repo:
            return next(
                (
                    pkg
                    for pkg in packages.values()
                    if pkg["package_or_url"] == self.package
                ),
                None,
            )
        else:
            return packages.get(self.package)

    @property
    def is_repo(self) -> bool:
        return self.package.startswith(GIT_PREFIX)

    async def local_version(self) -> Optional[str]:
        if self.is_repo:
            # pipx doesn't store Git commit, just the version field from the package
            return self.cached_version

        metadata = await self.get_metadata()
        if metadata:
            return metadata["package_version"]
        else:
            return None

    async def get_remote_version(self) -> str:
        if self.is_repo:
            return await repo_version(self.package)

        try:
            result = await self.driver.run_output(
                "python3", "-m", "pip", "index", "versions", self.package
            )
        except Exception as exc:
            raise Exception(
                f"Cannot find latest version of package '{self.package}'."
            ) from exc
        version = re.search(r"\(([^)]+)\)", result)
        if not version:
            raise Exception(f"Cannot parse pip output: {result}")  # pragma: no cover
        return version[1]

    async def install(self, *, tracker: Tracker) -> None:
        if self.is_repo:
            # https://github.com/pypa/pipx/issues/892
            cmd = ("install", "--force")
        else:
            cmd = ("upgrade", "--install")

        await self.driver.run("pipx", *cmd, self.package)

        tracker.track(await self.driver.local() / "pipx" / "venvs" / self.package)
        metadata = await self.get_metadata()
        if metadata:
            for bin_desc in metadata["app_paths"]:
                bin_path = Path(bin_desc["__Path__"]).name
                tracker.track(await self.driver.local() / "bin" / bin_path)

        await super().install(tracker=tracker)

        if self.is_repo:
            await self.cache_version()

    async def prerequisites(self) -> AsyncIterable[Package]:
        async for package in super().prerequisites():
            yield package  # pragma: no cover

        yield GitHubPackage(
            repo="pypa/pipx",
            binary=["pipx"],
            raw="pipx",
            db=self.db,
            driver=self.driver_,
        )

        os = await self.driver.os()

        for system in os.switch(
            linux=["python3-pip"],
            macos=[],
        ):
            yield SystemPackage(
                system=system,
                db=self.db,
                driver=self.driver_,
            )
